"""Functions use for testing."""
import fnmatch
import os
import shutil
from distutils.spawn import find_executable
from os.path import join
from pathlib import Path
from typing import Union

import h5py
import pkg_resources as pkg

__all__ = ["PATH_NANOQM", "PATH_TEST", "copy_basis_and_orbitals", "cp2k_available", "remove_files"]

# Environment data
PATH_NANOQM = Path(pkg.resource_filename('nanoqm', ''))
ROOT = PATH_NANOQM.parent
PATH_TEST = ROOT / "test" / "test_files"


def remove_files() -> None:
    """Remove tmp files in cwd."""
    for path in fnmatch.filter(os.listdir('.'), "plams_workdir*"):
        shutil.rmtree(path)
    for ext in ("hdf5", "db", "lock"):
        name = f"cache.{ext}"
        if os.path.exists(name):
            os.remove(name)


def cp2k_available(executable: str = "cp2k.popt") -> bool:
    """Check if cp2k is installed."""
    path = find_executable(executable)

    return path is not None



"""Test that the path are created propery."""

from pathlib import Path
from typing import Optional
import yaml

from nanoqm.common import UniqueSafeLoader
from nanoqm.workflows.initialization import initialize
from nanoqm.workflows.input_validation import process_input
from qmflows.type_hints import PathLike

from .utilsTest import PATH_TEST


def test_run_workflow(tmp_path: PathLike) -> None:
    """Check that all the paths are initialize."""
    create_config(tmp_path, True)
    create_config(tmp_path, False)


def create_config(tmp_path: str, scrath_is_None: bool) -> str:
    path = PATH_TEST / "input_fast_test_derivative_couplings.yml"
    with open(path, 'r') as f:
        inp = yaml.load(f, UniqueSafeLoader)

    # change scratch
    if scrath_is_None:
        inp["scratch_path"] = None
        del inp["path_hdf5"]
    else:
        inp["scratch_path"] = (tmp_path / "level0" / "level1").as_posix()
        # change HDF5 path
        inp["path_hdf5"] = (Path(inp["scratch_path"]) / "test_init.hdf5").as_posix()

    path_inp = tmp_path / "test_init.yml"
    with open(path_inp, 'w') as f:
        yaml.dump(inp, f)

    new_inp = process_input(path_inp, 'derivative_couplings')

    config = initialize(new_inp)

    assert Path(config.path_hdf5).exists()



"""Module to run restricted and unrestricted calculations."""

import logging
from typing import Any, Callable

from noodles import gather
from qmflows import run

from ..common import DictConfig
from .initialization import initialize

# Starting logger
logger = logging.getLogger(__name__)


def select_orbitals_type(
        config: DictConfig, workflow: Callable[[DictConfig], Any]) -> Any:
    """Call a workflow using restriced or unrestricted orbitals."""
    # Dictionary containing the general configuration
    config.update(initialize(config))

    if config.orbitals_type != "both":
        logger.info("starting workflow calculation!")
        promises = workflow(config)
        return run(promises, folder=config.workdir, always_cache=False)
    else:
        config_alphas = DictConfig(config.copy())
        config_betas = DictConfig(config.copy())
        config_alphas.orbitals_type = "alphas"
        promises_alphas = workflow(config_alphas)
        config_betas.orbitals_type = "betas"
        promises_betas = workflow(config_betas)
        all_promises = gather(promises_alphas, promises_betas)
        alphas, betas = run(all_promises, folder=config.workdir, always_cache=False)
        return alphas, betas



"""Test the analysis tools."""

from nanoqm.analysis.tools import parse_list_of_lists


def test_list_parser():
    """Check that a string representing a list of list is parsed correctly."""
    xs = '[[1,2,3,4]]'
    result = parse_list_of_lists(xs)
    assert result[0] == [1, 2, 3, 4]



"""Read basis in CP2K format."""
from pathlib import Path

import h5py
import pkg_resources
from qmflows.type_hints import PathLike

from nanoqm.workflows.initialization import store_cp2k_basis


def test_read_cp2k_basis(tmp_path: PathLike) -> None:
    """Read Basis set in CP2K format."""
    tmp_hdf5 = Path(tmp_path) / 'cp2k_basis.hdf5'
    tmp_hdf5.touch()

    path_basis = pkg_resources.resource_filename(
        "nanoqm", "basis/BASIS_MOLOPT")

    coefficients_format_carbon_DZVP_MOLOPT_GTH = '[2, 0, 2, 7, 2, 2, 1]'
    store_cp2k_basis(tmp_hdf5, path_basis)

    with h5py.File(tmp_hdf5, 'a') as f5:
        dset = f5["cp2k/basis/c/DZVP-MOLOPT-GTH/coefficients"]
        # Check that the format is store
        assert dset.attrs['basisFormat'] == coefficients_format_carbon_DZVP_MOLOPT_GTH
        # Check Shape of the coefficients
        assert dset.shape == (5, 7)



"""Check that the cp2k error is print to the user."""

from pathlib import Path

import pytest
from nanoqm.schedule.scheduleCP2K import try_to_read_wf


def test_cp2k_call_error(tmp_path: Path):
    """Check cp2k error files."""
    path_err = tmp_path / "cp2k.err"
    with open(path_err, 'w') as handler:
        handler.write("Some CP2K error")

    with pytest.raises(RuntimeError) as info:
        try_to_read_wf(tmp_path)

    error = info.value.args[0]
    print(error)
    assert "CP2K error" in error



"""Test the derivative coupling calculation."""
import os
import shutil
from typing import Sequence

import numpy as np

from assertionlib import assertion
from nanoqm.common import DictConfig, is_data_in_hdf5, retrieve_hdf5_data
from nanoqm.workflows.input_validation import process_input
from nanoqm.workflows.workflow_coupling import workflow_derivative_couplings

from .utilsTest import PATH_TEST, remove_files


def test_fast_couplings(tmp_path):
    """Check the derivative couplings workflow"""
    run_derivative_coupling(tmp_path, 'input_fast_test_derivative_couplings.yml')


def test_unrestricted_alphas(tmp_path):
    """Test the derivative coupling for the alphas spin orbitals."""
    run_derivative_coupling(tmp_path, 'input_couplings_alphas.yml', "alphas")


def test_unrestricted_betas(tmp_path):
    """Test the derivative coupling for the alphas spin orbitals."""
    run_derivative_coupling(tmp_path, 'input_couplings_both.yml', "both")


def run_derivative_coupling(tmp_path: str, input_file: str, orbitals_type: str = "") -> None:
    """Check that the couplings run."""
    path_input = PATH_TEST / input_file
    config = process_input(path_input, 'derivative_couplings')
    config["scratch_path"] = tmp_path
    tmp_hdf5 = os.path.join(tmp_path, 'fast_couplings.hdf5')
    shutil.copy(config.path_hdf5, tmp_hdf5)
    config['path_hdf5'] = tmp_hdf5
    config['write_overlaps'] = True
    try:
        check_results(config, tmp_hdf5, orbitals_type)
        # Run the calculation again to test that the data is read from the hdf5
        check_results(config, tmp_hdf5, orbitals_type)
    finally:
        remove_files()


def check_results(config: DictConfig, tmp_hdf5: str, orbitals_type: str) -> None:
    """Check the computed results stored in the HDF5 file."""
    if orbitals_type != "both":
        hamiltonians, _ = workflow_derivative_couplings(config)
        check_couplings(config, tmp_hdf5, orbitals_type)
        check_hamiltonians(hamiltonians)
    else:
        result_alphas, result_betas = workflow_derivative_couplings(config)
        check_couplings(config, tmp_hdf5, "alphas")
        check_couplings(config, tmp_hdf5, "betas")
        check_hamiltonians(result_alphas[0])
        check_hamiltonians(result_betas[0])


def check_couplings(config: DictConfig, tmp_hdf5: str, orbitals_type: str) -> None:
    """Check that the couplings have meaningful values."""
    def create_paths(keyword: str) -> list:
        return [os.path.join(orbitals_type, f'{keyword}_{x}')
                for x in range(len(config.geometries) - 1)]
    overlaps = create_paths('overlaps')
    couplings = create_paths('coupling')

    # Check that couplings and overlaps exists
    assertion.truth(is_data_in_hdf5(tmp_hdf5, overlaps))
    assertion.truth(is_data_in_hdf5(tmp_hdf5, couplings))

    # All the elements are different of inifinity or nan
    tensor_couplings = np.stack(retrieve_hdf5_data(tmp_hdf5, couplings))
    assertion.truth(np.isfinite(tensor_couplings).all())

    # Check that the couplings are anti-symetric
    for mtx in tensor_couplings[:]:
        assertion(np.allclose(mtx, -mtx.T))

    # Check that there are not NaN
    assertion.truth(not np.all(np.isnan(tensor_couplings)))


def check_hamiltonians(hamiltonians: Sequence[str]) -> None:
    """Check that the hamiltonians were written correctly."""
    energies = np.stack([np.diag(np.loadtxt(ts[1])) for ts in hamiltonians])
    couplings = np.stack([np.loadtxt(ts[0]) for ts in hamiltonians])

    # check that energies and couplings are finite values
    assertion.truth(np.isfinite(energies).all())
    assertion.truth(np.isfinite(couplings).all())

    # Check that the couplings diagonal is zero
    assertion.truth(abs(np.einsum('jii->', couplings)) < 1e-16)



"""Test the run_runworkflow script."""

import argparse
from pathlib import Path

import pytest
import yaml
from nanoqm.common import UniqueSafeLoader
from nanoqm.workflows.run_workflow import main
from pytest_mock import MockFixture

from .utilsTest import PATH_TEST


def call_main(mocker: MockFixture, path_input: Path, scratch_path: Path):
    """Mock main function."""
    # Mock argparse
    mocker.patch("argparse.ArgumentParser.parse_args", return_value=argparse.Namespace(
        i=path_input))

    mocker.patch("nanoqm.workflows.run_workflow.process_input", return_value={})
    mocker.patch("nanoqm.workflows.run_workflow.dict_workflows", return_value=len)
    main()


def test_run_workflow(mocker: MockFixture, tmp_path: Path):
    """Test that the CLI main command is called correctly."""
    path_input = PATH_TEST / "input_fast_test_derivative_couplings.yml"
    call_main(mocker, path_input, tmp_path)


def test_run_workflow_no_workflow(mocker: MockFixture, tmp_path: Path):
    """Check that an error is raised if not workflow is provided."""
    # remove workflow keyword
    with open(PATH_TEST / "input_fast_test_derivative_couplings.yml", 'r') as handler:
        input = yaml.load(handler, UniqueSafeLoader)
    input.pop('workflow')
    path_input = tmp_path / "wrong_input.yml"
    with open(path_input, 'w') as handler:
        yaml.dump(input, handler)

    with pytest.raises(RuntimeError) as info:
        call_main(mocker, path_input, tmp_path)

    error = info.value.args[0]
    assert "is required" in error



"""Test the distribution script."""
from qmflows.type_hints import PathLike
from pathlib import Path
from subprocess import (PIPE, Popen)
import fnmatch
import shutil
import os


def test_distribute(tmp_path: PathLike) -> None:
    """Check that the scripts to compute a trajectory are generated correctly."""
    cmd1 = "distribute_jobs.py -i test/test_files/input_test_distribute_derivative_couplings.yml"
    cmd2 = "distribute_jobs.py -i test/test_files/input_test_distribute_absorption_spectrum.yml"
    for cmd in [cmd1, cmd2]:
        call_distribute(tmp_path, cmd)


def call_distribute(tmp_path: PathLike, cmd: str) -> None:
    """Execute the distribute script and check that if finish succesfully."""
    try:
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
        _, err = p.communicate()
        if err:
            raise RuntimeError(err.decode())
        check_scripts()
    finally:
        remove_chunk_folder()


def check_scripts() -> None:
    """Check that the distribution scripts were created correctly."""
    paths = fnmatch.filter(os.listdir('.'), "chunk*")

    # Check that the files are created correctly
    files = ["launch.sh", "chunk_xyz*", "input.yml"]
    for p in paths:
        p = Path(p)
        for f in files:
            try:
                next(p.glob(f))
            except StopIteration:
                msg = f"There is not file: {f}"
                print(msg)
                raise RuntimeError(msg)


def remove_chunk_folder() -> None:
    """Remove resulting scripts."""
    for path in fnmatch.filter(os.listdir('.'), "chunk*"):
        shutil.rmtree(path)



"""Test input validation functionality."""
import pytest
from qmflows import cp2k, run
from qmflows.type_hints import PathLike
from scm import plams

from nanoqm.common import read_cell_parameters_as_array
from nanoqm.workflows.input_validation import process_input

from .utilsTest import PATH_TEST, cp2k_available, remove_files


def test_input_validation() -> None:
    """Test the templates and keywords completion."""
    path_input = PATH_TEST / "input_test_pbe0.yml"
    dict_input = process_input(path_input, "derivative_couplings")
    sett = dict_input['cp2k_general_settings']['cp2k_settings_guess']

    scale_x = sett.specific.cp2k.force_eval.dft.xc.xc_functional.pbe.scale_x

    assert abs(scale_x - 0.75) < 1e-16


@pytest.mark.skipif(
    not cp2k_available(), reason="CP2K is not install or not loaded")
def test_call_cp2k_pbe() -> None:
    """Check if the input for a PBE cp2k job is valid."""
    try:
        results = run_plams(PATH_TEST / "input_test_pbe.yml")
        assert (results is not None)
    finally:
        remove_files()


def run_plams(path_input: PathLike) -> float:
    """Call Plams to run a CP2K job."""
    # create settings
    dict_input = process_input(path_input, "derivative_couplings")
    sett = dict_input['cp2k_general_settings']['cp2k_settings_guess']

    # adjust the cell parameters
    file_cell_parameters = dict_input['cp2k_general_settings'].get("file_cell_parameters")
    if file_cell_parameters is not None:
        array_cell_parameters = read_cell_parameters_as_array(file_cell_parameters)[1]
        sett.cell_parameters = array_cell_parameters[0, 2:11].reshape(3, 3).tolist()

        # Run the job
    job = cp2k(sett, plams.Molecule(PATH_TEST / "C.xyz"))

    return run(job.energy)



"""Compute the excited states energies and MO coefficients using the STDDFT approach.

Index
-----
.. currentmodule:: nanoqm.workflows.workflow_stddft_spectrum
.. autosummary::
    workflow_stddft

"""
__all__ = ['workflow_stddft']

import logging
from os.path import join
from typing import Tuple

import numpy as np
from scipy.linalg import sqrtm
from scipy.spatial.distance import cdist
from noodles import gather, schedule, unpack
from noodles.interface import PromisedObject
from qmflows.parsers import parse_string_xyz
from qmflows.type_hints import PathLike

from ..common import (DictConfig, angs2au, change_mol_units, h2ev, hardness,
                      is_data_in_hdf5, number_spherical_functions_per_atom,
                      retrieve_hdf5_data, store_arrays_in_hdf5, xc)
from ..integrals.multipole_matrices import get_multipole_matrix
from ..schedule.components import calculate_mos
from .orbitals_type import select_orbitals_type

# Starting logger
logger = logging.getLogger(__name__)


def workflow_stddft(config: DictConfig) -> None:
    """Compute the excited states using simplified TDDFT.
    Both restricted and unrestricted orbitals calculations are available.

    Parameters
    ----------
    config
        Dictionary with the configuration to run the workflows

    """
    return select_orbitals_type(config, run_workflow_stddft)


def run_workflow_stddft(config: DictConfig) -> PromisedObject:
    """Compute the excited states using simplified TDDFT using `config`."""
    # Single Point calculations settings using CP2K
    mo_paths_hdf5, energy_paths_hdf5 = unpack(calculate_mos(config), 2)

    # Read structures
    molecules_au = [change_mol_units(parse_string_xyz(gs))
                    for i, gs in enumerate(config.geometries)
                    if (i % config.stride) == 0]

    # Noodles promised call
    scheduleTDDFT = schedule(compute_excited_states_tddft)

    results = gather(
        *[scheduleTDDFT(config, mo_paths_hdf5[i], DictConfig(
            {'i': i * config.stride, 'mol': mol}))
          for i, mol in enumerate(molecules_au)])

    return gather(results, energy_paths_hdf5)


def compute_excited_states_tddft(
        config: DictConfig, path_MOs: str, dict_input: DictConfig) -> None:
    """Compute the excited states properties (energy and coefficients).

    Take a given `mo_index_range`, the `tddft` method and `xc_dft` exchange functional.
    """
    logger.info("Reading energies and mo coefficients")
    # type of calculation
    energy, c_ao = retrieve_hdf5_data(config.path_hdf5, path_MOs)

    # Number of virtual orbitals
    nocc = config.active_space[0]
    nvirt = config.active_space[1]
    dict_input.update({"energy": energy, "c_ao": c_ao,
                       "nocc": nocc, "nvirt": nvirt})

    # Pass the molecule in Angstrom to the libint calculator
    copy_dict = DictConfig(dict_input.copy())
    copy_dict["mol"] = change_mol_units(dict_input["mol"], factor=1 / angs2au)

    # compute the multipoles if they are not stored
    multipoles = get_multipole_matrix(config, copy_dict, 'dipole')

    # read data from the HDF5 or calculate it on the fly
    dict_input["overlap"] = multipoles[0]

    # retrieve or compute the omega xia values
    omega, xia = get_omega_xia(config, dict_input)

    # add arrays to the dictionary
    dict_input.update(
        {"multipoles": multipoles[1:], "omega": omega, "xia": xia})

    compute_oscillator_strengths(config, dict_input)


def get_omega_xia(
        config: DictConfig, dict_input: DictConfig) -> Tuple[np.ndarray, np.ndarray]:
    """Search for the multipole_matrices, Omega and xia values in the HDF5.

    if they are not available compute and store them.

    Returns
    -------
    tuple
        omega and xia numpy arrays.

    """
    tddft = config.tddft.lower()

    def compute_omega_xia():
        if tddft == 'sing_orb':
            return compute_sing_orb(dict_input)

        return compute_std_aproximation(config, dict_input)

    # search data in HDF5
    point = f'point_{dict_input.i + config.enumerate_from}'
                
    paths_omega_xia = [join(x, point) for x in ("omega", "xia")]

    if is_data_in_hdf5(config.path_hdf5, paths_omega_xia):
        return tuple(retrieve_hdf5_data(config.path_hdf5, paths_omega_xia))
    
    else:
        omega, xia = compute_omega_xia()
        store_arrays_in_hdf5(
            config.path_hdf5, paths_omega_xia[0], omega, dtype=omega.dtype)
        store_arrays_in_hdf5(
            config.path_hdf5, paths_omega_xia[1], xia, dtype=xia.dtype)

        return omega, xia


def compute_sing_orb(inp: DictConfig) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the Single Orbital approximation."""
    energy, nocc, nvirt = tuple(inp[x]for x in ("energy", "nocc", "nvirt"))
    omega = -np.subtract(
        energy[:nocc].reshape(nocc, 1), energy[nocc:].reshape(nvirt, 1).T).reshape(nocc * nvirt)
    xia = np.eye(nocc * nvirt)

    return omega, xia


def compute_std_aproximation(
        config: DictConfig, dict_input: DictConfig) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the oscillator strenght using either the stda or stddft approximations."""
    logger.info("Reading or computing the dipole matrices")

    # Make a function tha returns in transition density charges
    logger.info("Computing the transition density charges")
    # multipoles[0] is the overlap matrix
    q = transition_density_charges(
        dict_input.mol, config, dict_input.overlap, dict_input.c_ao)

    # Make a function that compute the Mataga-Nishimoto-Ohno_Klopman
    # damped Columb and Excgange law functions
    logger.info(
        "Computing the gamma functions for Exchange and Coulomb integrals")
    gamma_J, gamma_K = compute_MNOK_integrals(dict_input["mol"], config.xc_dft)

    # Compute the Couloumb and Exchange integrals
    # If xc_dft is a pure functional, ax=0, thus the pqrs_J ints are not needed
    # and can be set to 0
    logger.info("Computing the Exchange and Coulomb integrals")
    if (xc(config.xc_dft)['type'] == 'pure'):
        size = dict_input.energy.size
        pqrs_J = np.zeros((size, size, size, size))
    else:
        pqrs_J = np.tensordot(q, np.tensordot(
            q, gamma_J, axes=(0, 1)), axes=(0, 2))
    pqrs_K = np.tensordot(q, np.tensordot(
        q, gamma_K, axes=(0, 1)), axes=(0, 2))

    # Construct the Tamm-Dancoff matrix A for each pair of i->a transition
    logger.info("Constructing the A matrix for TDDFT calculation")
    a_mat = construct_A_matrix_tddft(
        pqrs_J, pqrs_K, dict_input.nocc, dict_input.nvirt, config.xc_dft, dict_input.energy)

    if config.tddft == 'stddft':
        logger.info('sTDDFT has not been implemented yet !')
        # Solve the eigenvalue problem = A * cis = omega * cis
    elif config.tddft == 'stda':
        logger.info(
            "This is a TDA calculation ! \n Solving the eigenvalue problem")
        omega, xia = np.linalg.eig(a_mat)
    else:
        msg = f"The {config.tddft} method has not been implemented"
        raise NotImplementedError(msg)

    return omega, xia


def compute_oscillator_strengths(config: DictConfig, inp: DictConfig) -> None:
    """Compute oscillator strengths.

    The formula can be rearranged like this:
    f_I = 2/3 * np.sqrt(2 * omega_I) * sum_ia ( np.sqrt(e_diff_ia) * xia * tdm_x) ** 2 + y^2 + z^2
    """
    tddft = config.tddft.lower()
    # 1) Get the inp.energy matrix i->a. Size: Inp.Nocc * Inp.Nvirt
    delta_ia = -np.subtract(
        inp.energy[:inp.nocc].reshape(inp.nocc, 1),
        inp.energy[inp.nocc:].reshape(inp.nvirt, 1).T).reshape(inp.nocc * inp.nvirt)

    def compute_transition_matrix(matrix):
        if tddft == 'sing_orb':
            tm = np.stack(
                [np.sum(
                 delta_ia / inp.omega[i] * inp.xia[:, i] * matrix)
                 for i in range(inp.nocc * inp.nvirt)])
        else:
            tm = np.stack(
                [np.sum(
                 np.sqrt(2 * delta_ia / inp.omega[i]) * inp.xia[:, i] * matrix)
                 for i in range(inp.nocc * inp.nvirt)])
        return tm

    # 2) Compute the transition dipole matrix TDM(i->a)
    # Call the function that computes transition dipole moments integrals
    logger.info("Reading or computing the transition dipole matrix")

    def compute_tdmatrix(k):
        return np.linalg.multi_dot(
            [inp.c_ao[:, :inp.nocc].T, inp.multipoles[k, :, :],
             inp.c_ao[:, inp.nocc:]]).reshape(inp.nocc * inp.nvirt)

    td_matrices = (compute_tdmatrix(k) for k in range(3))

    # 3) Compute the transition dipole moments for each excited state i->a. Size: n_exc_states
    d_x, d_y, d_z = tuple(
        compute_transition_matrix(m) for m in td_matrices)

    # 4) Compute the oscillator strength
    f = 2 / 3 * inp.omega * (d_x ** 2 + d_y ** 2 + d_z ** 2)

    # Write to output
    inp.update({"dipole": (d_x, d_y, d_z), "oscillator": f})
    write_output(config, inp)


def write_output(config: DictConfig, inp: DictConfig) -> None:
    """Write the results using numpy functionality."""
    output = write_output_tddft(inp)

    path_output = join(config.workdir,
                       f'output_{inp.i + config.enumerate_from}_{config.tddft}.txt')
    fmt = '{:^5s}{:^14s}{:^8s}{:^11s}{:^11s}{:^11s}{:^11s}{:<5s}{:^10s}{:<5s}{:^11s}{:^11s}'
    header = fmt.format(
        'state', 'energy', 'f', 't_dip_x', 't_dip_y', 't_dip_y', 'weight',
        'from', 'energy', 'to', 'energy', 'delta_E')
    np.savetxt(path_output, output,
               fmt='%5d %10.3f %10.5f %10.5f %10.5f %10.5f %10.5f %3d %10.3f %3d %10.3f %10.3f',
               header=header)


def ex_descriptor(omega, f, xia, n_lowest, c_ao, s, tdm, tqm, nocc, nvirt, mol, config):
    """
    ADD DOCUMENTATION
    """
    # Reshape xia
    xia_I = xia.reshape(nocc, nvirt, nocc * nvirt)

    # Transform the transition density matrix into AO basis
    d0I_ao = np.stack(
        np.linalg.multi_dot(
            [c_ao[:, :nocc], xia_I[:, :, i], c_ao[:, nocc:].T]) for i in range(n_lowest))

    # Compute omega in excition analysis for the lowest n excitations
    om = get_omega(d0I_ao, s, n_lowest)

    # Compute the distribution of positions for the hole and electron
    xh, yh, zh = get_exciton_positions(d0I_ao, s, tdm, n_lowest, 'hole')
    xe, ye, ze = get_exciton_positions(d0I_ao, s, tdm, n_lowest, 'electron')

    # Compute the distribution of the square of position for the hole and electron
    x2h, y2h, z2h = get_exciton_positions(d0I_ao, s, tqm, n_lowest, 'hole')
    x2e, y2e, z2e = get_exciton_positions(d0I_ao, s, tqm, n_lowest, 'electron')

    # Compute the distribution of both hole and electron positions
    xhxe, yhye, zhze = get_exciton_positions(d0I_ao, s, tdm, n_lowest, 'both')

    # Compute Descriptors

    # Compute exciton size:
    d_exc = np.sqrt(
        ((x2h - 2 * xhxe + x2e) + (y2h - 2 * yhye + y2e) + (z2h - 2 * zhze + z2e)) / om)

    # Compute centroid electron_hole distance
    d_he = np.abs(((xe - xh) + (ye - yh) + (ze - zh)) / om)

    # Compute hole and electron size
    sigma_h = np.sqrt(
        (x2h / om - (xh / om) ** 2) + (y2h / om - (yh / om) ** 2) + (z2h / om - (zh / om) ** 2))
    sigma_e = np.sqrt(
        (x2e / om - (xe / om) ** 2) + (y2e / om - (ye / om) ** 2) + (z2e / om - (ze / om) ** 2))

    # Compute Pearson coefficients
    cov = (xhxe - xh * xe) + (yhye - yh * ye) + (zhze - zh * ze)
    r_eh = cov / (sigma_h * sigma_e)

    # Compute approximate d_exc and binding energy
    omega_ab = get_omega_ab(d0I_ao, s, n_lowest, mol, config)
    r_ab = get_r_ab(mol)

    d_exc_apprx = np.stack(
        np.sqrt(np.sum(omega_ab[i, :, :] * (r_ab ** 2)) / om[i]) for i in range(n_lowest))
    # binding energy approximated
    xs = np.stack((omega_ab[i, :, :] / r_ab) for i in range(n_lowest))
    xs[np.isinf(xs)] = 0
    binding_en_apprx = np.stack(
        (np.sum(xs[i, :, :]) / om[i]) for i in range(n_lowest))

    descriptors = write_output_descriptors(
        d_exc, d_exc_apprx, d_he, sigma_h, sigma_e, r_eh, binding_en_apprx, n_lowest, omega, f)

    return descriptors


def write_output_descriptors(
        d_exc, d_exc_apprx, d_he, sigma_h, sigma_e, r_eh, binding_ex_apprx, n_lowest, omega, f):
    """TODO: add Documentation."""
    au2ang = 0.529177249
    ex_output = np.empty((n_lowest, 10))
    ex_output[:, 0] = np.arange(n_lowest) + 1
    ex_output[:, 1] = d_exc * au2ang
    ex_output[:, 2] = d_exc_apprx * au2ang
    ex_output[:, 3] = d_he * au2ang
    ex_output[:, 4] = sigma_h * au2ang
    ex_output[:, 5] = sigma_e * au2ang
    ex_output[:, 6] = r_eh * au2ang
    ex_output[:, 7] = binding_ex_apprx * 27.211 / 2  # in eV
    ex_output[:, 8] = omega[:n_lowest] * 27.211
    ex_output[:, 9] = f[:n_lowest]

    return ex_output


def get_omega(d0I_ao, s, n_lowest):
    """TODO: add Documentation."""
    return np.stack(
        np.trace(
            np.linalg.multi_dot([d0I_ao[i, :, :].T, s, d0I_ao[i, :, :], s]))
        for i in range(n_lowest))


def get_r_ab(mol):
    """TODO: add Documentation."""
    coords = np.asarray([atom[1] for atom in mol])
    # Distance matrix between atoms A and B
    r_ab = cdist(coords, coords)
    return r_ab


def get_omega_ab(d0I_ao, s, n_lowest, mol, config):
    """TODO: add Documentation."""
    # Lowdin transformation of the transition density matrix
    n_atoms = len(mol)
    s_sqrt = sqrtm(s)
    d0I_mo = np.stack(
        np.linalg.multi_dot([s_sqrt, d0I_ao[i, :, :], s_sqrt]) for i in range(n_lowest))

    # Compute the number of spherical functions for each atom
    n_sph_atoms = number_spherical_functions_per_atom(
        mol, config['package_name'], config['basis_name'], config['path_hdf5'])

    # Compute omega_ab
    omega_ab = np.zeros((n_lowest, n_atoms, n_atoms))
    for i in range(n_lowest):
        index_a = 0
        for a in range(n_atoms):
            index_b = 0
            for b in range(n_atoms):
                omega_ab[i, a, b] = np.sum(
                    d0I_mo[i, index_a:(index_a + n_sph_atoms[a]),
                           index_b:(index_b + n_sph_atoms[b])] ** 2)
                index_b += n_sph_atoms[b]
            index_a += n_sph_atoms[a]

    return omega_ab


def get_exciton_positions(d0I_ao, s, moment, n_lowest, carrier):
    """TODO: add Documentation."""
    def compute_component_hole(k):
        return np.stack(
            np.trace(
                np.linalg.multi_dot([d0I_ao[i, :, :].T, moment[k, :, :], d0I_ao[i, :, :], s]))
            for i in range(n_lowest))

    def compute_component_electron(k):
        return np.stack(
            np.trace(
                np.linalg.multi_dot([d0I_ao[i, :, :].T, s, d0I_ao[i, :, :], moment[k, :, :]]))
            for i in range(n_lowest))

    def compute_component_he(k):
        return np.stack(
            np.trace(
                np.linalg.multi_dot(
                    [d0I_ao[i, :, :].T, moment[0, :, :], d0I_ao[i, :, :], moment[0, :, :]]))
            for i in range(n_lowest))

    if carrier == 'hole':
        return tuple(compute_component_hole(k) for k in range(3))
    elif carrier == 'electron':
        return tuple(compute_component_electron(k) for k in range(3))
    elif carrier == 'both':
        return tuple(compute_component_he(k) for k in range(3))
    else:
        raise RuntimeError(f"unkown option: {carrier}")


def write_output_tddft(inp: DictConfig) -> np.ndarray:
    """Write out as a table in plane text."""
    energy = inp.energy

    excs = [(i, a) for i in range(inp.nocc)
            for a in range(inp.nocc, inp.nvirt + inp.nocc)]

    output = np.empty((inp.nocc * inp.nvirt, 12))
    output[:, 0] = 0  # State number: we update it after reorder
    output[:, 1] = inp.omega * h2ev  # State energy in eV
    output[:, 2] = inp.oscillator  # Oscillator strength

    d_x, d_y, d_z = inp.dipole
    output[:, 3] = d_x  # Transition dipole moment in the x direction
    output[:, 4] = d_y  # Transition dipole moment in the y direction
    output[:, 5] = d_z  # Transition dipole moment in the z direction
    # Weight of the most important excitation
    output[:, 6] = np.hstack([np.max(inp.xia[:, i] ** 2)
                              for i in range(inp.nocc * inp.nvirt)])

    # Find the index of this transition
    index_weight = np.hstack([
        np.where(
            inp.xia[:, i] ** 2 == np.max(
                inp.xia[:, i] ** 2))
        for i in range(inp.nocc * inp.nvirt)]).reshape(inp.nocc * inp.nvirt)

    # Index of the hole for the most important excitation
    output[:, 7] = np.stack([excs[index_weight[i]][0]
                             for i in range(inp.nocc * inp.nvirt)]) + 1
    # These are the energies of the hole for the transition with the larger weight
    output[:, 8] = energy[output[:, 7].astype(int) - 1] * h2ev
    # Index of the electron for the most important excitation
    output[:, 9] = np.stack([excs[index_weight[i]][1]
                             for i in range(inp.nocc * inp.nvirt)]) + 1
    # These are the energies of the electron for the transition with the larger weight
    output[:, 10] = energy[output[:, 9].astype(int) - 1] * h2ev
    # This is the energy for the transition with the larger weight
    output[:, 11] = (
        energy[output[:, 9].astype(int) - 1] - energy[output[:, 7].astype(int) - 1]) * h2ev

    # Reorder the output in ascending order of energy
    output = output[output[:, 1].argsort()]
    # Give a state number in the correct order
    output[:, 0] = np.arange(inp.nocc * inp.nvirt) + 1

    return output


def transition_density_charges(mol, config, s, c_ao):
    """TODO: add Documentation."""
    n_atoms = len(mol)
    sqrt_s = sqrtm(s)
    c_mo = np.dot(sqrt_s, c_ao)
    # Size of the transition density tensor : n_atoms x n_mos x n_mos
    q = np.zeros((n_atoms, c_mo.shape[1], c_mo.shape[1]))
    n_sph_atoms = number_spherical_functions_per_atom(
        mol, config['package_name'], config.cp2k_general_settings['basis'], config['path_hdf5'])

    index = 0
    for i in range(n_atoms):
        q[i, :, :] = np.dot(
            c_mo[index:(index + n_sph_atoms[i]), :].T, c_mo[index:(index + n_sph_atoms[i]), :])
        index += n_sph_atoms[i]

    return q


def compute_MNOK_integrals(mol, xc_dft):
    """TODO: add Documentation."""
    n_atoms = len(mol)
    r_ab = get_r_ab(mol)
    hardness_vec = np.stack([hardness(m[0]) for m in mol]).reshape(n_atoms, 1)
    print("hardness: ", [m[0].capitalize() for m in mol])
    hard = np.add(hardness_vec, hardness_vec.T)
    beta = xc(xc_dft)['beta1'] + xc(xc_dft)['ax'] * xc(xc_dft)['beta2']
    alpha = xc(xc_dft)['alpha1'] + xc(xc_dft)['ax'] * xc(xc_dft)['alpha2']
    if (xc(xc_dft)['type'] == 'pure'):
        gamma_J = np.zeros((n_atoms, n_atoms))
    else:
        gamma_J = np.power(
            1 / (np.power(r_ab, beta) + np.power((xc(xc_dft)['ax'] * hard), -beta)), 1 / beta)
    gamma_K = np.power(1 / (np.power(r_ab, alpha) +
                            np.power(hard, -alpha)), 1 / alpha)

    return gamma_J, gamma_K


def construct_A_matrix_tddft(pqrs_J, pqrs_K, nocc, nvirt, xc_dft, e):
    """TODO: add Documentation."""
    # This is the exchange integral entering the A matrix.
    #  It is in the format (nocc, nvirt, nocc, nvirt)
    k_iajb = pqrs_K[:nocc, nocc:, :nocc, nocc:].reshape(
        nocc * nvirt, nocc * nvirt)

    # This is the Coulomb integral entering in the A matrix.
    # It is in the format: (nocc, nocc, nvirt, nvirt)
    k_ijab_tmp = pqrs_J[:nocc, :nocc, nocc:, nocc:]

    # To get the correct order in the A matrix, i.e. (nocc, nvirt, nocc, nvirt),
    # we have to swap axes
    k_ijab = np.swapaxes(k_ijab_tmp, axis1=1, axis2=2).reshape(
        nocc * nvirt, nocc * nvirt)

    # They are in the m x m format where m is the number of excitations = nocc * nvirt
    a_mat = 2 * k_iajb - k_ijab

    # Generate a vector with all possible ea - ei energy differences
    e_diff = -np.subtract(
        e[:nocc].reshape(nocc, 1), e[nocc:].reshape(nvirt, 1).T).reshape(nocc * nvirt)
    np.fill_diagonal(a_mat, np.diag(a_mat) + e_diff)

    return a_mat



"""Test he absorption spectrum workflows."""
import shutil
from os.path import join
from pathlib import Path

import numpy as np
from nanoqm.common import retrieve_hdf5_data
from nanoqm.workflows import workflow_stddft
from nanoqm.workflows.input_validation import process_input

from .utilsTest import PATH_TEST, remove_files


def run_and_check(tmp_path: Path, project: str, input_file: str, orbitals_type: str = ""):
    """Run the workflow and check the results."""
    path_original_hdf5 = PATH_TEST / f'{project}.hdf5'
    # create scratch path
    shutil.copy(path_original_hdf5, tmp_path)
    for approx in ("sing_orb", "stda"):
        try:
            # Run the actual test
            path_test_hdf5 = Path(tmp_path) / f"{project}_{approx}.hdf5"
            shutil.copyfile(path_original_hdf5, path_test_hdf5)
            calculate_oscillators(path_test_hdf5, tmp_path, approx, input_file)
            check_properties(path_test_hdf5, orbitals_type)

            # Run again the workflow to check that the data is read from the hdf5
            calculate_oscillators(path_test_hdf5, tmp_path, approx, input_file)
            check_properties(path_test_hdf5, orbitals_type)
        finally:
            remove_files()


def test_compute_oscillators(tmp_path: Path):
    """Compute the oscillator strenght and check the results."""
    # path_original_hdf5 = PATH_TEST / 'Cd.hdf5'
    input_file = 'input_test_absorption_spectrum.yml'
    run_and_check(tmp_path, "Cd", input_file)


def test_compute_oscillators_unrestricted(tmp_path: Path):
    """Compute the oscillator strenght and check the results."""
    input_file = 'input_test_absorption_spectrum_unrestricted.yml'
    run_and_check(tmp_path, "oxygen", input_file, "alphas")


def calculate_oscillators(path_test_hdf5: Path, scratch_path: Path, approx: str, file_name: str):
    """Compute a couple of couplings with the Levine algorithm using precalculated MOs."""
    input_file = PATH_TEST / file_name
    config = process_input(input_file, 'absorption_spectrum')
    config['path_hdf5'] = path_test_hdf5.absolute().as_posix()
    config['scratch_path'] = scratch_path
    config['workdir'] = scratch_path
    config['tddft'] = approx
    config['path_traj_xyz'] = Path(config.path_traj_xyz).absolute().as_posix()

    workflow_stddft(config)


def check_properties(path_test_hdf5: Path, orbitals_type: str):
    """Check that the tensor stored in the HDF5 are correct."""
    path_dipole = join(orbitals_type, 'dipole', 'point_0')
    dipole_matrices = retrieve_hdf5_data(
        path_test_hdf5, path_dipole)

    # The diagonals of each component of the matrix must be zero
    # for a single atom
    diagonals = np.sum([np.diag(dipole_matrices[n + 1]) for n in range(3)])
    assert abs(diagonals) < 1e-16



"""Test quadrupole calculation."""

import shutil
from pathlib import Path

import numpy as np
from assertionlib import assertion
from nanoqm.integrals.multipole_matrices import compute_matrix_multipole
from nanoqm.workflows.input_validation import process_input
from qmflows.parsers.xyzParser import readXYZ

from .utilsTest import PATH_TEST


def test_quadropole(tmp_path):
    """Test the calculation of a quadrupole."""
    file_path = PATH_TEST / "input_test_single_points.yml"
    config = process_input(file_path, 'single_points')
    path_original_hdf5 = config.path_hdf5
    path_test_hdf5 = (Path(tmp_path) / "multipoles.hdf5").as_posix()

    # copy the precomputed data to the temporal HDF5
    shutil.copyfile(path_original_hdf5, path_test_hdf5)
    config.path_hdf5 = path_test_hdf5

    mol = readXYZ((PATH_TEST / "ethylene.xyz").as_posix())
    matrix = compute_matrix_multipole(mol, config, "quadrupole")
    # The matrix contains the overlap + dipole + quadrupole
    assertion.shape_eq(matrix, (10, 46, 46))
    # Check that the matrices are symmetric
    for i in range(10):
        arr = matrix[i].reshape(46, 46)
        assertion.truth(np.allclose(arr, arr.T))



import re
from unittest import mock

import setuptools
import pkg_resources
from assertionlib import assertion

PATTERN = re.compile("@ ")


def test_requirements_txt() -> None:
    """Test that ``test_requirements.txt`` and ``setup.py`` are synced."""
    with open("test_requirements.txt", "r") as f:
        requirements_set = {PATTERN.sub("@", str(i)) for i in pkg_resources.parse_requirements(f)}

    with mock.patch.object(setuptools, "setup") as mock_setup:
        import setup
    _, kwargs = mock_setup.call_args

    setup_set = set(kwargs["install_requires"])
    setup_set.update(kwargs["extras_require"]["test"])

    diff_set = requirements_set ^ setup_set
    assertion.not_(diff_set, message="Dependency mismatch between setup.py and test_requirements.txt")



"""Tests using mypy."""

import subprocess
import warnings

from qmflows.test_utils import Assertion_Warning

from typing_extensions import Literal

from .utilsTest import PATH_NANOQM, PATH_TEST

PACKAGE = PATH_NANOQM.absolute().as_posix()
INI = (PATH_TEST / 'mypy.ini').absolute().as_posix()

Action = Literal['raise', 'warn', 'ignore']  #: Type annotation for the 'action' keyword.
ACTION = frozenset(['raise', 'warn', 'ignore'])


def test_mypy(action: Action = 'warn') -> None:
    """Test using mypy."""
    if action not in ACTION:
        raise ValueError(f"Invalid value for the 'action' parameter: {action!r}")

    command = f"mypy {PACKAGE!r} --config-file {INI!r}"
    out = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    stdout = out.stdout.decode()
    stderr = out.stderr.decode()
    assert not stderr, stderr

    if action == 'warn' and out.returncode != 0:
        warnings.warn(stdout, category=Assertion_Warning)

    elif action == 'raise':
        try:
            assert out.returncode == 0, stdout
        except AssertionError as ex:
            msg = stdout.rsplit('\n', maxsplit=2)[1]
            raise AssertionError(msg) from ex



"""Test a single point calculation using CP2K."""
import os
from pathlib import Path

import pytest
from assertionlib import assertion

from nanoqm.common import is_data_in_hdf5
from nanoqm.workflows.input_validation import process_input
from nanoqm.workflows.workflow_single_points import workflow_single_points

from .utilsTest import PATH_TEST, cp2k_available, remove_files


def run_single_point(tmp_path: Path, input_file: str):
    """Run a single point calculation using cp2k."""
    file_path = PATH_TEST / input_file
    config = process_input(file_path, 'single_points')
    config["scratch_path"] = tmp_path
    tmp_hdf5 = os.path.join(tmp_path, 'single_points.hdf5')
    config['path_hdf5'] = tmp_hdf5
    Path(tmp_hdf5).touch()
    try:
        path_orbitals, path_energies = workflow_single_points(config)
        print("path_orbitals: ", path_orbitals)
        print("path_energies: ", path_energies)
        if config["compute_orbitals"]:
            assertion.truth(is_data_in_hdf5(tmp_hdf5, path_orbitals[0]))
    finally:
        remove_files()


# @pytest.mark.skipif(
#     not cp2k_available(), reason="CP2K is not install or not loaded")
@pytest.mark.slow
def test_single_point(tmp_path: Path):
    """Check that the couplings run."""
    run_single_point(tmp_path, "input_test_single_points.yml")



"""Installation recipe."""
import os
import sys
from os.path import join
from typing import Dict

import setuptools
from Cython.Distutils import build_ext
from setuptools import Extension, find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

version = {}  # type: Dict[str, str]
with open(os.path.join(here, 'nanoqm', '__version__.py')) as f:
    exec(f.read(), version)


def readme():
    """Load readme."""
    with open('README.rst') as f:
        return f.read()


class get_pybind_include:
    """Helper class to determine the pybind11 include path.

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked.
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on the specified compiler.

    As of Python 3.6, CCompiler has a `has_flag` method.
    http: // bugs.python.org/issue26689
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cc') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[17/14/11] compiler flag.

    The newer version is prefered over c++11 (when it is available).
    """
    flags = ['-std=c++17', '-std=c++14', '-std=c++11']

    for flag in flags:
        if has_flag(compiler, flag):
            return flag

    raise RuntimeError('Unsupported compiler -- at least C++11 support '
                       'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }
    l_opts = {
        'msvc': [],
        'unix': [],
    }

    if sys.platform == 'darwin':
        darwin_opts = ['-stdlib=libc++', '-mmacosx-version-min=10.14', '-fno-sized-deallocation']
        c_opts['unix'] += darwin_opts
        l_opts['unix'] += darwin_opts

    def build_extensions(self):
        """Actual compilation."""
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        link_opts = self.l_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' %
                        self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' %
                        self.distribution.get_version())
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = link_opts
        build_ext.build_extensions(self)


# Set path to the conda libraries
conda_prefix = os.environ["CONDA_PREFIX"]
if conda_prefix is None:
    raise RuntimeError(
        "No conda module found. A Conda environment is required")

conda_include = join(conda_prefix, 'include')
conda_lib = join(conda_prefix, 'lib')
ext_pybind = Extension(
    'compute_integrals',
    sources=['libint/compute_integrals.cc'],
    include_dirs=[
        # Path to pybind11 headers
        'libint/include',
        conda_include,
        join(conda_include, 'eigen3'),
        get_pybind_include(),
        get_pybind_include(user=True),
        '/usr/include/eigen3'
    ],
    libraries=['hdf5', 'int2'],
    library_dirs=[conda_lib],
    language='c++')

setup(
    name='nano-qmflows',
    version=version['__version__'],
    description='Derivative coupling calculation',
    license='Apache-2.0',
    url='https://github.com/SCM-NV/nano-qmflows',
    author=['Felipe Zapata', 'Ivan Infante'],
    author_email='f.zapata@esciencecenter.nl',
    keywords='chemistry Photochemistry Simulation',
    long_description=readme() + '\n\n',
    long_description_content_type='text/markdown',
    packages=find_packages(),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Typing :: Typed',
    ],
    install_requires=[
        'h5py',
        'mendeleev',
        'more-itertools',
        'noodles==0.3.3',
        'numpy',
        'scipy',
        'schema',
        'pyyaml>=5.1',
        'plams>=1.5.1',
        'qmflows>=0.11.0',
    ],
    cmdclass={'build_ext': BuildExt},
    ext_modules=[ext_pybind],
    extras_require={
        'test': [
            'assertionlib',
            'codacy-coverage',
            'mypy',
            'pytest',
            'pytest-cov',
            'pytest-mock',
            'pytest-pycodestyle',
            'pytest-pydocstyle',
            'types-pyyaml',
            'types-pkg_resources',
            'Cython',
            'setuptools',
        ],
        'doc': ['sphinx>=2.1', 'sphinx-autodoc-typehints', 'sphinx_rtd_theme', 'nbsphinx']
    },
    include_package_data=True,
    package_data={
        'nanoqm': ['basis/*.json', 'basis/BASIS*', 'basis/GTH_POTENTIALS', 'py.typed']
    },
    entry_points={
        'console_scripts': [
            'run_workflow.py=nanoqm.workflows.run_workflow:main',
            'distribute_jobs.py=nanoqm.workflows.distribute_jobs:main'
        ]
    },
    scripts=[
        'scripts/convert_legacy_hdf5.py',
        'scripts/hamiltonians/plot_mos_energies.py',
        'scripts/hamiltonians/plot_spectra.py',
        'scripts/pyxaid/plot_average_energy.py',
        'scripts/pyxaid/plot_cooling.py',
        'scripts/pyxaid/plot_spectra_pyxaid.py',
        'scripts/pyxaid/plot_states_pops.py',
        'scripts/qmflows/mergeHDF5.py',
        'scripts/qmflows/plot_dos.py',
        'scripts/qmflows/rdf.py',
        'scripts/qmflows/removeHDF5folders.py',
        'scripts/qmflows/remove_mos_hdf5.py',
        'scripts/qmflows/convolution.py'
    ]
)



"""Test the workflows tools."""
import numpy as np
from qmflows.parsers import parse_string_xyz

from nanoqm.common import number_spherical_functions_per_atom

from .utilsTest import PATH_TEST


def test_calc_sphericals():
    """Test the calculation of spherical functions."""
    with open(PATH_TEST / 'Cd33Se33.xyz', 'r') as f:
        mol = parse_string_xyz(f.read())
    path_hdf5 = PATH_TEST / "Cd33Se33.hdf5"
    xs = number_spherical_functions_per_atom(
        mol, "cp2k", "DZVP-MOLOPT-SR-GTH", path_hdf5)

    expected = np.concatenate((np.repeat(25, 33), np.repeat(13, 33)))

    assert np.array_equal(xs, expected)



"""Check the schemas."""
from assertionlib import assertion

from nanoqm.workflows.input_validation import process_input

from .utilsTest import PATH_TEST


def test_input_validation():
    """Test the input validation schema."""
    schemas = ("absorption_spectrum", "derivative_couplings")
    paths = [PATH_TEST / x for x in
             ["input_test_absorption_spectrum.yml", "input_fast_test_derivative_couplings.yml"]]
    for s, p in zip(schemas, paths):
        d = process_input(p, s)
        assertion.isinstance(d, dict)



"""Test the IPR workflow."""
import os
import shutil
import sys
from os.path import join

from qmflows.type_hints import PathLike

from nanoqm.workflows.input_validation import process_input
from nanoqm.workflows.workflow_ipr import workflow_ipr

from .utilsTest import PATH_TEST


def test_workflow_IPR(tmp_path: PathLike) -> None:
    """Test the Inverse Participation Ratio workflow."""
    file_path = PATH_TEST / 'input_test_IPR.yml'
    config = process_input(file_path, 'ipr_calculation')
    # create scratch path
    shutil.copy(config.path_hdf5, tmp_path)
    config.path_hdf5 = join(tmp_path, "Cd33Se33.hdf5")
    config.workdir = tmp_path

    try:
        workflow_ipr(config)
        os.remove("IPR.txt")
    except BaseException:
        print("scratch_path: ", tmp_path)
        print("Unexpected error:", sys.exc_info()[0])
        raise
